import time
from IPython import display
from IPython.display import HTML, Markdown
from ipywidgets import Output

from rubix.aws.base import get_client
from rubix.utils import print_markdown_table, first, printmd


def get_latest_deployment_status(service, **kwargs):
    client = get_client('ecs', **kwargs)
    cluster = kwargs.get('cluster')

    if cluster and not isinstance(cluster, str):
        raise ValueError('Cluster value should be of type str')

    response = client.describe_services(
        cluster=cluster,
        services=[service]
    )

    if not response['services']:
        raise ValueError('Service could not be found. Please check service name, cluster name, and aws region. Make sure the AWS credentials have rights to access ECS!')

    if not response['services'][0]['deployments']:
        print('No deployments found! Have you just spin up the service recently?')

    deployments = response['services'][0]['deployments']
    recent_deployment = {}

    for deployment in deployments:
        if deployment['status'] == 'PRIMARY':
            recent_deployment = deployment.copy()

    recent_deployment.pop('status', None)
    recent_deployment.pop('launchType', None)
    return recent_deployment


def rollback_deployment(service, **kwargs):
    client = get_client('ecs', **kwargs)
    cluster = kwargs.get('cluster')

    if cluster and not isinstance(cluster, str):
        raise ValueError('Cluster value should be of type str')

    if cluster:
        try:
            response = client.describe_services(
                cluster=cluster,
                services=[service]
            )
        except client.exceptions.ServiceNotFoundException:
            raise ValueError('Could not find service %s. Please check service, cluster, region and that IAM credentials has access to ECS' % service)
    else:
        try:
            # default cluster is assumed when not specified
            response = client.describe_services(
                services=[service]
            )
        except client.exceptions.ClusterNotFoundException:
            raise ValueError('Could not find service %s in default cluster. Please specify cluster.' % service)

    if not response['services']:
        raise ValueError('Service could not be found. Please check service name, cluster name, and aws region. Make sure the AWS credentials have rights to access ECS!')

    current_task_definiton = response['services'][0]['taskDefinition']
    print('Current task definition for service %s is:: %s' % (service, current_task_definiton))

    response = client.describe_task_definition(
        taskDefinition=current_task_definiton
    )

    task_definiton_family = response['taskDefinition']['family']

    response = client.list_task_definitions(
        familyPrefix=task_definiton_family,
        status='INACTIVE',
        sort='DESC',
        maxResults=5
    )

    task_definitions = response['taskDefinitionArns']

    if not task_definitions:
        raise ValueError('Could not find previous task definitions for service %s. Make sure that the previous task definiton exist and they are marked INACTIVE.' % service)

    rollback_candidates = []

    for td in task_definitions:
        response = client.describe_task_definition(
            taskDefinition=td
        )

        image = response['taskDefinition']['containerDefinitions'][0]['image']
        rollback_candidates.append([td, image])

    print_markdown_table(headers=['Task Definitions', 'Container Image'], data=rollback_candidates)
    rollback_td = input('Enter the task defintion that you want to rollback to: ')

    response = client.describe_task_definition(
        taskDefinition=rollback_td
    )

    response = response['taskDefinition']

    new_td = client.register_task_definition(
        family=response['family'],
        containerDefinitions=response['containerDefinitions']
    )

    client.update_service(
        cluster=cluster,
        service=service,
        taskDefinition=new_td['taskDefinition']['taskDefinitionArn']
    )

    markdown_string = ''
    progress_bar_url = "http://progressed.io/bar/0"

    deployment_status_output = Output()
    display.display(deployment_status_output)

    # poll for 10 minutes for updates (120 * 5 second wait)
    for i in range(121):
        response = client.describe_services(
            cluster=cluster,
            services=[service]
        )

        deployments = response['services'][0]['deployments']
        primary_deployment = first(d for d in deployments if d['status'] == 'PRIMARY')

        desired_count = primary_deployment['desiredCount']
        running_count = primary_deployment['runningCount']

        status = int(running_count * 100 / desired_count)
        progress_bar_url = "http://progressed.io/bar/" + str(status)
        markdown_string = "# New Task Definition\n" + "* **Desired Count: " + str(desired_count) + "**\n" + "* **Running Count: " + str(running_count) + "**"
        deployment_status_output.clear_output(wait=True)

        with deployment_status_output:
            printmd(markdown_string)
            display.display(
                HTML(
                    "<table><tr><td><b>Deployment Status</b></td><td><img src=" + progress_bar_url + "></td></tr></table>"))
            display.display(HTML("<div class=\"loader\"></div>"))

        if status == 100:
            break

        time.sleep(4.5)

    deployment_status_output.clear_output(wait=True)

    with deployment_status_output:
        printmd(markdown_string)
        display.display(
            HTML(
                "<table><tr><td><b>Deployment Status</b></td><td><img src=" + progress_bar_url + "></td></tr></table>"))
        display.display(HTML("<div class=\"stoppedloader\"></div>"))



import os
import boto3


def get_client(service, **kwargs):
    '''
    Validates credentials and region is present.
    Returns client for the required AWS service.
    '''

    aws_access_key_id = kwargs.get('aws_access_key_id') or os.environ.get('AWS_ACCESS_KEY_ID')
    aws_secret_access_key = kwargs.get('aws_secret_access_key') or os.environ.get('AWS_SECRET_ACCESS_KEY')
    aws_region = kwargs.get('aws_region') or os.environ.get('AWS_REGION')

    if not aws_access_key_id:
        raise ValueError(
            'aws_access_key_id is not specified. Either send it as a method argument: aws_access_key_id Or set the environment variable: AWS_ACCESS_KEY_ID')

    if not aws_secret_access_key:
        raise ValueError(
            'aws_secret_access_key is not specified. Either send it as a method argument: aws_secret_access_key Or set the environment variable: AWS_SECRET_ACCESS_KEY')

    if not aws_region:
        raise ValueError(
            'aws_region is not specified. Either send it as a method argument: aws_region Or set the environment variable: AWS_REGION')

    return boto3.client(
        service,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=aws_region
    )



from IPython.display import HTML, display, Markdown


def print_html_table(data):
    '''
    Takes in a list of list as data and prints an HTML table
    '''
    display(HTML(
        '<table><tr>{}</tr></table>'.format(
            '</tr><tr>'.join(
                '<td>{}</td>'.format('</td><td>'.join(str(_) for _ in row)) for row in data)
            )
     ))


def print_markdown_table(headers, data):
    '''
    Renders table given headers and data
    '''
    md = ''

    for h in headers:
        md += '|' + h

    md += '|\n'

    for r in range(len(headers)):
        md += '|---'

    md += '|\n'

    for row in data:
        for element in row:
            md += '|' + str(element)
        md += '|\n'

    display(Markdown(md))


def first(iterable, default=None):
    '''
    Returns the first item from iterable. If no items then returns default
    '''
    for item in iterable:
      return item

    return default


def printmd(string):
    '''
    Renders markdown in notebook output of codecell
    '''
    display(Markdown(string))



from kubernetes import config, client


def get_latest_deployment_status(service_name, namespace='default', context=None):

    try:
        config.load_kube_config(context=context)
    except FileNotFoundError:
        raise RuntimeError("Could not find Kubernetes config. "
                           "Have you uploaded the config file from Nurtch admin UI? "
                           "If recently uploaded please wait couple of minutes for the file to propagate to every node.")

    core_client = client.CoreV1Api()
    response = core_client.read_namespaced_service(name=service_name, namespace=namespace)
    selector = response.spec.selector

    if not selector:
        raise RuntimeError('This service has no label selector. '
                           'Services without label selector are not handled in Rubix. Use kubectl instead.')

    selector_string = ''

    for k, v in selector.items():
        selector_string += (k + '=' + v + ',')

    selector_string = selector_string[:-1]  # remove trailing comma

    apps_client = client.AppsV1Api()
    response = apps_client.list_replica_set_for_all_namespaces(label_selector=selector_string)

    if response and response.items:
        replica_sets = response.items
        latest_rs = replica_sets[0]
        latest_rs_created_at = replica_sets[0].metadata.creation_timestamp

        for current_rs in replica_sets:
            current_rs_created_at = current_rs.metadata.creation_timestamp

            if current_rs_created_at > latest_rs_created_at:
                latest_rs = current_rs
                latest_rs_created_at = current_rs_created_at

        result = {}
        result['desiredCount'] = latest_rs.spec.replicas
        result['availableCount'] = latest_rs.status.available_replicas
        result['currentCount'] = latest_rs.status.replicas
        result['containerImage'] = latest_rs.spec.template.spec.containers[0].image
        result['createdAt'] = latest_rs.metadata.creation_timestamp

        return result
    else:
        raise RuntimeError('Did not find any recent deployment history of the requested service!')


import plotly
import pandas as pd
import plotly.graph_objs as go

from collections import OrderedDict
from datetime import datetime, timezone, timedelta

from .base import get_client

POSSIBLE_STATISTICS = {'SampleCount', 'Average', 'Sum', 'Minimum', 'Maximum'}


def plot_metric(namespace, metric_name, **kwargs):
    client = get_client('cloudwatch', **kwargs)
    statistics = 'Average'  # default stat we'll be using for plotting
    markers = kwargs.get('markers', [])

    if kwargs.get('start_time') and not isinstance(kwargs.get('start_time'), datetime):
        raise ValueError('start_time argument should be of type datetime.datetime')

    if kwargs.get('end_time') and not isinstance(kwargs.get('end_time'), datetime):
        raise ValueError('end_time argument should be of type datetime.datetime')

    if kwargs.get('statistics'):
        statistics = kwargs.get('statistics')

        if not isinstance(statistics, str):
            raise ValueError('statistics argument should be a string')

        if statistics not in POSSIBLE_STATISTICS:
            raise ValueError('Invalid statistics argument. statistics can only take following values: %s', POSSIBLE_STATISTICS)

    if kwargs.get('dimensions') and not isinstance(kwargs.get('dimensions'), list):
        raise ValueError('Cloudwatch Metric Dimensions should be specificed as a list.')

    if not isinstance(markers, list):
        raise ValueError('Markers should be of type list')

    dimensions = kwargs.get('dimensions') or []
    utc_dt = datetime.now(timezone.utc)  # UTC time
    default_end_time = utc_dt.astimezone()
    end_time = kwargs.get('end_time') or default_end_time
    default_start_time = end_time - timedelta(seconds=3600 * 12)
    start_time = kwargs.get('start_time') or default_start_time

    for marker in markers:

        if not isinstance(marker, datetime):
            raise ValueError('Marker %s should be of type datetime.datetime' % marker)

    # ignore markers not falling between start and end time
    markers = [marker for marker in markers if (start_time <= marker <= end_time)]

    response = client.get_metric_statistics(
        Namespace=namespace,
        MetricName=metric_name,
        StartTime=start_time,
        EndTime=end_time,
        Dimensions=dimensions,
        Period=300,
        Statistics=[statistics],
    )
    ds = response['Datapoints']

    if not ds:
        raise ValueError('No metric data received from cloudwatch. Make sure that the resource exist. Check namespace, metric name, dimensions, region.')

    title = namespace + ' ' + metric_name
    ds2 = {}
    max_timestamp = ds[0]['Timestamp']
    max_value = ds[0][statistics]

    for d in ds:
        new_val = d[statistics]
        ds2[d['Timestamp']] = new_val

        if new_val > max_value:
            max_timestamp = d['Timestamp']
            max_value = new_val

    ds3 = OrderedDict(sorted(ds2.items(), key=lambda t: t[0]))
    ds4 = pd.DataFrame.from_dict(ds3, orient="index")
    ds4.index.name = 'timestamp'
    ds4.columns = ['count']

    plotly.offline.init_notebook_mode(connected=True)
    data = [go.Scatter(x=ds4.index.tolist(), y=ds4['count'].tolist(), name=title)]

    # let's draw marker lines
    shapes = []

    for marker in markers:
        shapes.append({
            'type': 'line',
            'x0': marker,
            'y0': 0,
            'x1': marker,
            'y1': max_value,
            'line': {
                'color': 'rgb(50, 171, 96)',
                'width': 4,
                'dash': 'dashdot',
            }
        })

    layout = {
        'title': title,
        'shapes': shapes
    }

    plotly.offline.iplot({
        "data": data,
        "layout": layout,
    })



import ez_setup
ez_setup.use_setuptools()

from setuptools import setup, find_packages

setup(
    name='rubix',
    version='0.0.8',
    author='Amit Rathi',
    description='Python package to enables DevOps tasks in Jupyter Notebooks',
    long_description=open('README.md').read(),
    license='GNU Lesser General Public License v3.0',
    keywords=['DevOps', 'Jupyter', 'nurtch'],
    packages=find_packages(),
    package_data={'rubix': ['rubix/assets/css/*.css']},
    include_package_data=True,
    install_requires=[
        'wheel',
        'plotly',
        'pandas',
        'boto3',
        'kubernetes',
        'ipywidgets'
    ]
)